/*
 * tnef.c -- extract files from microsoft TNEF format
 *
 * Copyright (C)1999, 2000 Mark Simpson <damned@world.std.com>
 * Copyright (C)1998 Thomas Boll  <tb@boll.ch>		[ORIGINAL AUTHOR]
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, you can either send email to this
 * program's maintainer or write to: The Free Software Foundation,
 * Inc.; 59 Temple Place, Suite 330; Boston, MA 02111-1307, USA.
 *
 * Commentary:
 *       scans tnef file and extracts all attachments
 *       attachments are written to their original file-names if possible
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "strdup.h"

#include "tnef.h"

/* To quiet compiler define tempnam */
extern char*
tempnam(const char*, const char*);

/* Needed to transform char buffers into little endian numbers */
#define GETINT32(p)    (uint32)((p)[0]+((p)[1]<<8)+((p)[2]<<16)+((p)[3]<<24))
#define GETINT16(p)    (uint16)((p)[0]+((p)[1]<<8))

/* Program Options: */
static int Verbose_Output = 0;  /* Shall we shout it from the rooftops? */
static int Dry_Run = 0;         /* Parse file only */
static FILE* Input_File = NULL; /* Input File */
static char* Directory = NULL;  /* Where to put output */


/* Array of days of the week for translating a date */
static char* day_of_week[] = { "Sun", "Mon", "Tue",
                               "Wed", "Thu", "Fri", "Sat" }; 

/* Format Strings for dumping unsigned integers */
#if (SIZEOF_INT == 4)
#define SHORT_INT_FMT "%hu"
#define LONG_INT_FMT  "%u"
#else
#define SHORT_INT_FMT "%u"
#define LONG_INT_FMT  "%ul"
#endif

/* Attr -- storing a structure, formated according to file specification */
typedef struct 
{
    uint16 type;
    uint16 name;
    unsigned long len;
    char* buf;
} Attr;


/* ********************
   UTILITY FUNCTIONS
   ******************** */
/* attempts to malloc memory, if fails print error and call abort */
static void*
Malloc (size_t size)
{
    void *ptr = malloc (size);
    if (!ptr)
    {
        perror ("Memory allocation failure");
        abort();
    }
    return ptr;
}

/* concatenates fname1 with fname2 to make a pathname, adds '/' as needed
 */ 
static char *
concat_fname (const char *fname1, const char* fname2)
{
    char *filename = NULL;
    int len = 0;
    if (fname1) len += strlen (fname1);
    if (fname2) len += strlen (fname2);
    if (len == 0) return NULL;
    
    filename = (char*)Malloc ((len + 2) * sizeof (char));
    
    strcpy (filename, fname1);
    if (fname1[strlen(fname1)-1] != '/')
        strcat (filename, "/");
    strcat (filename, fname2);

    return filename;
}


/* 
   geti32, geti16, geti8:
   Get 8, 16 or 32 bits from the file 
*/
static uint32
geti32 ()
{
    unsigned char buf[4];

    if (fread (buf, 4, 1, Input_File) != 1) 
    {
        fprintf (stderr, "unexpected end of input\n");
        exit (1);
    }
    return (uint32)GETINT32(buf);
}

static uint16
geti16 ()
{
    unsigned char buf[2];

    if (fread (buf, 2, 1, Input_File) != 1) 
    {
        fprintf (stderr, "unexpected end of input\n");
        exit (1);
    }
    return (uint16)GETINT16(buf);
}

/* dump_attr
   print attr to stderr.  Assumes that the Verbose flag has been set and
   already chedked */
static void 
dump_attr (Attr* attr)
{
    char *name = get_attr_name_str (attr->name);
    char *type = get_attr_type_str (attr->type);
    struct date dt;
    uint16 s;
    uint32 l;
    unsigned long i;

    fprintf (stderr, "%s [type: %s] =", name, type);

    switch (attr->type)
    {
    case szBYTE:
        for (i=0; i < attr->len; i+=1) 
        {
            if (i< 10) fprintf (stderr, " %02x", (uint8)attr->buf[i]);
            else if (i==10) fprintf (stderr, "...");
        }
        break;

    case szSHORT:
        if (attr->len < sizeof(uint16))
        {
            fprintf (stderr, "Not enough data for szSHORT");
            abort();
        }
		s = GETINT16(attr->buf);
        fprintf (stderr, " " SHORT_INT_FMT, s);
        if (attr->len > sizeof(uint16))
        {
            fprintf (stderr, " [extra data:");
            for (i = sizeof(uint16); i < attr->len; i++)
            {
                fprintf (stderr, " %02x", (uint8)attr->buf[i]);
            }
            fprintf (stderr, " ]");
        }
        break;

    case szLONG:
        if (attr->len < sizeof(uint32))
        {
            fprintf (stderr, "Not enough data for szLONG");
            abort();
        }
		l = GETINT32(attr->buf);
        fprintf (stderr, " " LONG_INT_FMT, l);
        if (attr->len > sizeof(uint32))
        {
            fprintf (stderr, " [extra data:");
            for (i = sizeof(uint32); i < attr->len; i++)
            {
                fprintf (stderr, " %02x", (uint8)attr->buf[i]);
            }
            fprintf (stderr, " ]");
        }
        break;
        

    case szWORD:
        for (i=0; i < attr->len; i+=2) 
        {
            if (i < 6) fprintf (stderr, " %04x", GETINT16(attr->buf));
            else if (i==6) fprintf (stderr, "...");
        }
        break;

    case szDWORD:
        for (i=0; i < attr->len; i+=4) 
        {
            if (i < 4) fprintf (stderr, " %08x", GETINT32(attr->buf));
            else if (i==4) fprintf (stderr, "...");
        }
        break;

    case szDATE:
        memmove (&dt, attr->buf, attr->len);
		dt.dow = GETINT16((unsigned char *)&dt.dow);
		dt.year = GETINT16((unsigned char *)&dt.year); 
		dt.month = GETINT16((unsigned char *)&dt.month);
		dt.day = GETINT16((unsigned char *)&dt.day);
		dt.hour = GETINT16((unsigned char *)&dt.hour);
		dt.min = GETINT16((unsigned char *)&dt.min);
		dt.sec = GETINT16((unsigned char *)&dt.sec);
        fprintf (stderr, " %s %04d/%02d/%02d %02d:%02d:%02d",
                 day_of_week[dt.dow],
                 dt.year, dt.month, dt.day,
                 dt.hour, dt.min, dt.sec);
        break;

    case szTEXT:
    case szSTRING:
        {
            char* buf = Malloc ((attr->len + 1) * sizeof (char));
            strncpy (buf, (char*)attr->buf, attr->len);
            buf[attr->len] = '\0';
            fprintf (stderr, "'%s'", buf);
            free (buf);
        }
        break;

    default:
        fprintf (stderr, "<unknown type> len=%ld", attr->len);
    }
    fprintf (stderr, "\n");
}


/* Validate the checksum against attr.  The checksum is the sum of all the
   bytes in the attribute data modulo 65536 */
static int
check_checksum (Attr* attr, uint16 checksum)
{
    uint32 i;
    uint16 sum = 0;

    for (i = 0; i < attr->len; i++)
    {
        sum += (uint8)attr->buf[i];
    }
    sum %= 65536;
    return (sum == checksum);
}


/* takss int d and decodes the attribute from it, returns the attr, this
   functions reads from the file */
static Attr*
decode_attribute (unsigned int d) 
{
    Attr* attr = (Attr*)calloc (1, sizeof (Attr));
    unsigned long bytes_read = 0;
    uint16 checksum;

    assert (attr);
    if (!attr) abort();

    attr->type = (d >> 16);
    attr->name = ((d << 16) >> 16);
    attr->len = geti32();
    attr->buf = Malloc (attr->len);

    bytes_read = fread (attr->buf, 1, attr->len, Input_File);
    if (bytes_read != attr->len)
    {
        fprintf (stderr, "Unexpected end of input!\n");
        exit (1);
    }
    
    checksum = geti16();
    if (!check_checksum(attr, checksum))
    {
        fprintf (stderr, 
                 "Invalid checksum, input file may be corrupted\n");
        exit (1);
    }
    
    if (Verbose_Output) dump_attr (attr);

    return attr;
}


static void
decode_message (void) 
{  
    uint32 d;
    Attr *attr = NULL;
    d = geti32();
    attr = decode_attribute(d);
    free (attr);
}



static void
decode_attachment (void) 
{  
    char *buf = NULL;
    uint32 d;
    FILE *fp;
    static char *fname = NULL;
    static char *prevtmpfname = NULL;
    Attr *attr = NULL;
    static unsigned long file_counter = 1;
    
    d = geti32();
    attr = decode_attribute (d);
    if (!attr) abort();
    
    switch (attr->name)
    {
    case attATTACHTITLE:
        /* check to see if we have a filename... if not we will make a
           default and pretend that we actually got it */
        if (attr->len && *attr->buf == '\0')
        {
            if (Verbose_Output)
            {
                fprintf (stderr, 
                         "No file name specified, using default.\n");
            }
            attr->len = 32;     /* seems like a reasonable size */
            attr->buf = Malloc ((attr->len + 1) * sizeof (char));
            sprintf (attr->buf, "tnef-tmp.%ld", file_counter++);
        }
        buf = strdup (attr->buf);
        if (fname) free(fname);
        fname = concat_fname (Directory, buf);
        free (buf);

        /* here is where we rename the file if the filename came after the
           data */ 
        if (prevtmpfname)
        {
            fprintf (stderr, "%sRENAMING %s to %s\n", 
                 ((Dry_Run==0)?"":"NOT "), prevtmpfname, fname);
            if (Dry_Run == 0)
            {
                if(rename(prevtmpfname, fname) == -1)
                {
                    perror(prevtmpfname);
                }
            }
            if (fname) free(fname);
            fname=NULL;
            if (prevtmpfname) free(prevtmpfname);
            prevtmpfname=NULL;
        }
        break;

    case attATTACHDATA:
        if (fname == NULL)
        {
            buf=malloc((32+1) * sizeof (char));
        	sprintf (buf, "tnef-tmp.%ld", file_counter++);
        	fname = concat_fname (Directory, buf);
        	free (buf);

            /* since the fname was not available, save the tmp filename so
               we can rename the prevtmpfname to the real filename if we
               receive it next in attATTACHTITLE */
            prevtmpfname = strdup (fname);
        }

        fprintf (stderr, "%sWRITING %s\n", 
                 ((Dry_Run==0)?"":"NOT "), fname);
        if (Dry_Run == 0)
        {
            fp = fopen (fname, "wb");
            if (fp == NULL) 
            {
                perror (fname);
                exit (1);
            }
            fwrite (attr->buf, attr->len, 1, fp);
            fclose (fp);
        }

        free (fname); fname = NULL;
        break;
  
    default:
        break;
    }
    free (attr);
}


int
parse_file (FILE* input_file, char* directory, int verbose, int dry_run)
{
    uint32 d;
    uint16 key;

    /* store the program options in our file global variables */
    Input_File = input_file;
    Directory = directory;
    Verbose_Output = verbose;
    Dry_Run = dry_run;

    /* check that this is in fact a TNEF file */
    d = geti32();
    if (d != TNEF_SIGNATURE) 
    {
        fprintf (stderr, "Seems not to be a TNEF file\n");
        return 1;
    }

    /* Get the key */
    key = geti16();
    if (verbose) fprintf (stderr, "TNEF Key is: %hx\n", key);

    /* The rest of the file is a series of 'messages' and 'attachments' */
    for (;;) 
    {
        char buf[1];
        uint8 c;

        if (fread (buf, 1, 1, Input_File) == 0) break;
        c = (uint8)buf[0];

        switch (c) 
        {
        case LVL_MESSAGE:
            decode_message();
            break;
        case LVL_ATTACHMENT:
            decode_attachment();
            break;
        default:
            fprintf (stderr, "Coding Error in TNEF file\n");
            return 1;
        }
    }
    return 0;
}
